@@ -5,6 +5,7 @@ use crate::python_interpreter::InterpreterKind;
 use crate::BuildContext;
 use crate::CargoToml;
 use crate::Metadata21;
+use crate::PyProjectToml;
 use crate::PythonInterpreter;
 use crate::Target;
 use anyhow::{bail, format_err, Context, Result};
@@ -105,10 +106,12 @@ impl Default for BuildOptions {
 
 impl BuildOptions {
     /// Tries to fill the missing metadata for a BuildContext by querying cargo and python
-    pub fn into_build_context(self, release: bool, strip: bool) -> Result<BuildContext> {
-        if self.platform_tag == Some(PlatformTag::manylinux1()) {
-            eprintln!("⚠  Warning: manylinux1 is unsupported by the Rust compiler.");
-        }
+    pub fn into_build_context(
+        self,
+        release: bool,
+        strip: bool,
+        editable: bool,
+    ) -> Result<BuildContext> {
         let manifest_file = &self.manifest_path;
         if !manifest_file.exists() {
             let current_dir =
@@ -130,6 +133,16 @@ impl BuildOptions {
 
         let cargo_toml = CargoToml::from_path(&manifest_file)?;
         let manifest_dir = manifest_file.parent().unwrap();
+        let pyproject: Option<PyProjectToml> = if manifest_dir.join("pyproject.toml").is_file() {
+            let pyproject =
+                PyProjectToml::new(manifest_dir).context("pyproject.toml is invalid")?;
+            pyproject.warn_missing_maturin_version();
+            pyproject.warn_missing_build_backend();
+            Some(pyproject)
+        } else {
+            None
+        };
+        let pyproject = pyproject.as_ref();
         let metadata21 = Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir)
             .context("Failed to parse Cargo.toml into python metadata")?;
         let extra_metadata = cargo_toml.remaining_core_metadata();
@@ -142,7 +155,7 @@ impl BuildOptions {
             .lib
             .as_ref()
             .and_then(|lib| lib.name.as_ref())
-            .unwrap_or(&crate_name)
+            .unwrap_or(crate_name)
             .to_owned();
 
         // Only use extension name from extra metadata if it contains dot
@@ -152,9 +165,22 @@ impl BuildOptions {
             .filter(|name| name.contains('.'))
             .unwrap_or(&module_name);
 
-        let project_layout = ProjectLayout::determine(manifest_dir, &extension_name)?;
+        let project_layout = ProjectLayout::determine(
+            manifest_dir,
+            extension_name,
+            extra_metadata.python_source.as_deref(),
+        )?;
 
-        let mut cargo_extra_args = split_extra_args(&self.cargo_extra_args)?;
+        let mut args_from_pyproject = Vec::new();
+        let mut cargo_extra_args = self.cargo_extra_args.clone();
+        if cargo_extra_args.is_empty() {
+            // if not supplied on command line, try pyproject.toml
+            if let Some(args) = pyproject.and_then(|x| x.cargo_extra_args()) {
+                cargo_extra_args.push(args.to_string());
+                args_from_pyproject.push("cargo-extra-args");
+            }
+        }
+        cargo_extra_args = split_extra_args(&cargo_extra_args)?;
         if let Some(ref target) = self.target {
             cargo_extra_args.extend(vec!["--target".to_string(), target.clone()]);
         }
@@ -179,7 +205,17 @@ impl BuildOptions {
             }
         };
 
-        let bridge = find_bridge(&cargo_metadata, self.bindings.as_deref())?;
+        let bridge = find_bridge(
+            &cargo_metadata,
+            self.bindings.as_deref().or_else(|| {
+                pyproject.and_then(|x| {
+                    if x.bindings().is_some() {
+                        args_from_pyproject.push("bindings");
+                    }
+                    x.bindings()
+                })
+            }),
+        )?;
 
         if bridge != BridgeModel::Bin && module_name.contains('-') {
             bail!(
@@ -204,7 +240,15 @@ impl BuildOptions {
             None => find_interpreter(&bridge, &[], &target, get_min_python_minor(&metadata21))?,
         };
 
-        let rustc_extra_args = split_extra_args(&self.rustc_extra_args)?;
+        let mut rustc_extra_args = self.rustc_extra_args.clone();
+        if rustc_extra_args.is_empty() {
+            // if not supplied on command line, try pyproject.toml
+            if let Some(args) = pyproject.and_then(|x| x.rustc_extra_args()) {
+                rustc_extra_args.push(args.to_string());
+                args_from_pyproject.push("rustc-extra-args");
+            }
+        }
+        rustc_extra_args = split_extra_args(&rustc_extra_args)?;
 
         let mut universal2 = self.universal2;
         // Also try to determine universal2 from ARCHFLAGS environment variable
@@ -224,6 +268,27 @@ impl BuildOptions {
                 universal2 = true;
             }
         };
+        let strip = pyproject.map(|x| x.strip()).unwrap_or_default() || strip;
+        let skip_auditwheel =
+            pyproject.map(|x| x.skip_auditwheel()).unwrap_or_default() || self.skip_auditwheel;
+        let platform_tag = self.platform_tag.or_else(|| {
+            pyproject.and_then(|x| {
+                if x.compatibility().is_some() {
+                    args_from_pyproject.push("compatibility");
+                }
+                x.compatibility()
+            })
+        });
+        if platform_tag == Some(PlatformTag::manylinux1()) {
+            eprintln!("⚠️  Warning: manylinux1 is unsupported by the Rust compiler.");
+        }
+
+        if !args_from_pyproject.is_empty() {
+            eprintln!(
+                "📡 Using build options {} from pyproject.toml",
+                args_from_pyproject.join(", ")
+            );
+        }
 
         Ok(BuildContext {
             target,
@@ -236,13 +301,14 @@ impl BuildOptions {
             out: wheel_dir,
             release,
             strip,
-            skip_auditwheel: self.skip_auditwheel,
-            platform_tag: self.platform_tag,
+            skip_auditwheel,
+            platform_tag,
             cargo_extra_args,
             rustc_extra_args,
             interpreter,
             cargo_metadata,
             universal2,
+            editable,
         })
     }
 }
@@ -252,14 +318,14 @@ impl BuildOptions {
 fn get_min_python_minor(metadata21: &Metadata21) -> Option<usize> {
     if let Some(requires_python) = &metadata21.requires_python {
         let regex = Regex::new(r#">=3\.(\d+)(?:\.\d)?"#).unwrap();
-        if let Some(captures) = regex.captures(&requires_python) {
+        if let Some(captures) = regex.captures(requires_python) {
             let min_python_minor = captures[1]
                 .parse::<usize>()
                 .expect("Regex must only match usize");
             Some(min_python_minor)
         } else {
             println!(
-                "⚠  Couldn't parse the value of requires-python, \
+                "⚠️ Couldn't parse the value of requires-python, \
                     not taking it into account when searching for python interpreter. \
                     Note: Only `>=3.x.y` is currently supported."
             );
@@ -374,23 +440,23 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
         if !pyo3_node.features.contains(&"extension-module".to_string()) {
             let version = cargo_metadata[&pyo3_node.id].version.to_string();
             println!(
-                "⚠  Warning: You're building a library without activating pyo3's \
+                "⚠️  Warning: You're building a library without activating pyo3's \
                  `extension-module` feature. \
                  See https://pyo3.rs/v{}/building_and_distribution.html#linking",
                 version
             );
         }
 
-        if let Some((major, minor)) = has_abi3(&cargo_metadata)? {
+        return if let Some((major, minor)) = has_abi3(cargo_metadata)? {
             println!(
                 "🔗 Found pyo3 bindings with abi3 support for Python ≥ {}.{}",
                 major, minor
             );
-            return Ok(BridgeModel::BindingsAbi3(major, minor));
+            Ok(BridgeModel::BindingsAbi3(major, minor))
         } else {
             println!("🔗 Found pyo3 bindings");
-            return Ok(bridge);
-        }
+            Ok(bridge)
+        };
     }
 
     Ok(bridge)
@@ -416,7 +482,7 @@ fn find_single_python_interpreter(
         );
     };
 
-    let interpreter = PythonInterpreter::check_executable(executable, &target, &bridge)
+    let interpreter = PythonInterpreter::check_executable(executable, target, bridge)
         .context(format_err!(err_message))?
         .ok_or_else(|| format_err!(err_message))?;
     Ok(interpreter)
@@ -434,10 +500,10 @@ pub fn find_interpreter(
     match bridge {
         BridgeModel::Bindings(binding_name) => {
             let mut interpreter = if !interpreter.is_empty() {
-                PythonInterpreter::check_executables(&interpreter, &target, &bridge)
+                PythonInterpreter::check_executables(interpreter, target, bridge)
                     .context("The given list of python interpreters is invalid")?
             } else {
-                PythonInterpreter::find_all(&target, &bridge, min_python_minor)
+                PythonInterpreter::find_all(target, bridge, min_python_minor)
                     .context("Finding python interpreters failed")?
             };
 
@@ -451,7 +517,7 @@ pub fn find_interpreter(
 
             if binding_name == "pyo3" && target.is_unix() && is_cross_compiling(target)? {
                 if let Some(cross_lib_dir) = std::env::var_os("PYO3_CROSS_LIB_DIR") {
-                    println!("⚠ Cross-compiling is poorly supported");
+                    println!("⚠️ Cross-compiling is poorly supported");
                     let host_python = &interpreter[0];
                     println!(
                         "🐍 Using host {} for cross-compiling preparation",
@@ -463,8 +529,12 @@ pub fn find_interpreter(
                     env::set_var("PYTHON_SYS_EXECUTABLE", &host_python.executable);
 
                     let sysconfig_path = find_sysconfigdata(cross_lib_dir.as_ref(), target)?;
-                    let sysconfig_data =
-                        parse_sysconfigdata(&host_python.executable, sysconfig_path)?;
+                    env::set_var(
+                        "MATURIN_PYTHON_SYSCONFIGDATA_DIR",
+                        sysconfig_path.parent().unwrap(),
+                    );
+
+                    let sysconfig_data = parse_sysconfigdata(host_python, sysconfig_path)?;
                     let major = sysconfig_data
                         .get("version_major")
                         .context("version_major is not defined")?
@@ -507,6 +577,8 @@ pub fn find_interpreter(
                         interpreter_kind,
                         abi_tag,
                         libs_dir: PathBuf::from(cross_lib_dir),
+                        platform: None,
+                        runnable: false,
                     }];
                 }
             }
@@ -529,6 +601,12 @@ pub fn find_interpreter(
         }
         BridgeModel::Bin => Ok(vec![]),
         BridgeModel::BindingsAbi3(major, minor) => {
+            let interpreter = if !interpreter.is_empty() {
+                PythonInterpreter::check_executables(interpreter, target, bridge)
+                    .unwrap_or_default()
+            } else {
+                PythonInterpreter::find_all(target, bridge, min_python_minor).unwrap_or_default()
+            };
             // Ideally, we wouldn't want to use any python interpreter without abi3 at all.
             // Unfortunately, on windows we need one to figure out base_prefix for a linker
             // argument.
@@ -536,7 +614,7 @@ pub fn find_interpreter(
                 if let Some(manual_base_prefix) = std::env::var_os("PYO3_CROSS_LIB_DIR") {
                     // PYO3_CROSS_LIB_DIR should point to the `libs` directory inside base_prefix
                     // when cross compiling, so we fake a python interpreter matching it
-                    println!("⚠ Cross-compiling is poorly supported");
+                    println!("⚠️  Cross-compiling is poorly supported");
                     Ok(vec![PythonInterpreter {
                         major: *major as usize,
                         minor: *minor as usize,
@@ -547,20 +625,19 @@ pub fn find_interpreter(
                         interpreter_kind: InterpreterKind::CPython,
                         abi_tag: None,
                         libs_dir: PathBuf::from(manual_base_prefix),
+                        platform: None,
+                        runnable: false,
                     }])
                 } else {
-                    let interpreter = find_single_python_interpreter(
-                        bridge,
-                        interpreter,
-                        target,
-                        "abi3 on windows",
-                    )?;
-                    println!("🐍 Using {} to generate to link bindings (With abi3, an interpreter is only required on windows)", interpreter);
-                    Ok(vec![interpreter])
+                    let interp = interpreter
+                        .get(0)
+                        .context("Failed to find a python interpreter")?;
+                    println!("🐍 Using {} to generate to link bindings (With abi3, an interpreter is only required on windows)", interp);
+                    Ok(interpreter)
                 }
             } else {
                 println!("🐍 Not using a specific python interpreter (With abi3, an interpreter is only required on windows)");
-                Ok(vec![])
+                Ok(interpreter)
             }
         }
     }
@@ -570,7 +647,7 @@ pub fn find_interpreter(
 fn split_extra_args(given_args: &[String]) -> Result<Vec<String>> {
     let mut splitted_args = vec![];
     for arg in given_args {
-        match shlex::split(&arg) {
+        match shlex::split(arg) {
             Some(split) => splitted_args.extend(split),
             None => {
                 bail!(
@@ -730,7 +807,7 @@ mod test {
         let mut options = BuildOptions::default();
         options.cargo_extra_args.push("--features log".to_string());
         options.bindings = Some("bin".to_string());
-        let context = options.into_build_context(false, false).unwrap();
+        let context = options.into_build_context(false, false, false).unwrap();
         assert_eq!(context.cargo_extra_args, vec!["--features", "log"])
     }
 
